# -*- encoding: utf-8 -*-
"""
keri.vc.walleting module

"""
from .. import help
from ..core.coring import Counter, CtrDex, Prefixer, Seqner, Diger, Siger
from ..core.parsing import Parser, Colds
from ..core.scheming import JSONSchema, CacheResolver
from ..db import dbing
from ..kering import ShortageError, ColdStartError, ExtractionError, UnverifiedProofError
from ..vc.proving import Credentialer

# TODO: create this and populate with needed schema for now
cache = CacheResolver()

logger = help.ogler.getLogger()


def openPocket(name="test", **kwa):
    """
    Returns contextmanager generated by openLMDB but with Baser instance
    """
    return dbing.openLMDB(cls=Pocketer, name=name, **kwa)


def parseCredential(ims=b'', wallet=None, typ=JSONSchema()):
    """
    Parse the ims bytearray as a CESR Proof Format verifiable credential

    Parameters:
        ims (bytearray) of serialized incoming verifiable credential in CESR Proof Format.
        wallet (Wallet) storage for the verified credential
        typ (JSONSchema) class for resolving schema references:

    """
    try:
        creder = Credentialer(raw=ims, typ=typ)
    except ShortageError as e:
        raise e
    else:
        del ims[:creder.size]
    cold = Parser.sniff(ims)
    if cold is Colds.msg:
        raise ColdStartError("unable to parse VC, attachments expected")

    ctr = Parser.extract(ims=ims, klas=Counter, cold=cold)
    if ctr.code != CtrDex.AttachedMaterialQuadlets:
        raise ExtractionError("Invalid attachment to VC {}, expected {}"
                              "".format(ctr.code, CtrDex.AttachedMaterialQuadlets))

    pags = ctr.count * 4
    if len(ims) != pags:
        raise ShortageError("VC proof attachment invalid length {}, expected {}"
                            "".format(len(ims), pags))

    prefixer, seqner, diger, isigers = parseProof(ims)

    wallet.processCredential(creder, prefixer, seqner, diger, isigers)


def parseProof(ims=b''):
    cold = Parser.sniff(ims)
    if cold is Colds.msg:
        raise ColdStartError("unable to parse VC, attachments expected")

    ctr = Parser.extract(ims=ims, klas=Counter, cold=cold)
    if ctr.code != CtrDex.TransIndexedSigGroups or ctr.count != 1:
        raise ExtractionError("Invalid attachment to VC {}, expected one {}"
                              "".format(ctr.code, CtrDex.TransIndexedSigGroups))

    prefixer = Parser.extract(ims=ims, klas=Prefixer)
    seqner = Parser.extract(ims=ims, klas=Seqner)
    diger = Parser.extract(ims=ims, klas=Diger)

    ictr = Parser.extract(ims=ims, klas=Counter)
    if ictr.code != CtrDex.ControllerIdxSigs:
        raise ExtractionError("Invalid attachment to VC {}, expected {}"
                              "".format(ctr.code, CtrDex.ControllerIdxSigs))

    isigers = []
    for i in range(ictr.count):
        isiger = Parser.extract(ims=ims, klas=Siger)
        isigers.append(isiger)

    return prefixer, seqner, diger, isigers


def buildProof(prefixer, seqner, diger, sigers):
    """
    
    Parameters:
        prefixer (Prefixer) Identifier of the issuer of the credential
        seqner (Seqner) is the sequence number of the event used to sign the credential
        diger (Diger) is the digest of the event used to sign the credential
        sigers (list) are the cryptographic signatures on the credential
    
    """

    prf = bytearray()
    prf.extend(Counter(CtrDex.TransIndexedSigGroups, count=1).qb64b)
    prf.extend(prefixer.qb64b)
    prf.extend(seqner.qb64b)
    prf.extend(diger.qb64b)

    prf.extend(Counter(code=CtrDex.ControllerIdxSigs, count=len(sigers)).qb64b)
    for siger in sigers:
        prf.extend(siger.qb64b)

    return prf


class Pocketer(dbing.LMDBer):
    """
    Pocketer is the store for the wallet (get it?).


        .issus is named subDB of issuer prefix identifiers to Credential SAIDs
            representing the credentials issued by an Issuer
            DB is keyed by identifer prefix of the Issuer of the credential
            More than one value per DB key is allowed
        .subjs is named subDB of subject prefix identifiers to Credential SAIDs
            representing the credentials issued TO the subject identified by the prefix
            DB is keyed by identifer prefix of the subject of the credential
            More than one value per DB key is allowed
        .sers is the named subDB of raw serialized bytes of the Credential.  Represents
            what was signed
            key is Credential SAID
            Only one value per DB key is allowed
        .seals is the named subDB of the Event Location Seal triple of pre+snu+dig
            of the location in the KEL where the VC was signed
            key is Credential SAID
            Only one value per DB key is allowed
        .sigs is named sub DB of event proof quadruples from transferable
            signers. Each quadruple is concatenation of  four fully qualified items
            of validator. These are: transferable prefix, plus latest establishment
            event sequence number plus latest establishment event digest,
            plus indexed event signature.
            When latest establishment event is multisig then there will
            be multiple quadruples one per signing key, each a dup at same db key.
            dgKey
            DB is keyed by Credential SAID
            More than one value per DB key is allowed


    """
    TailDirPath = "keri/pck"
    AltTailDirPath = ".keri/pck"
    TempPrefix = "keri_pck_"

    def __init__(self, headDirPath=None, reopen=True, **kwa):
        """
        Setup named sub databases.

        Parameters:
            name is str directory path name differentiator for main database
                When system employs more than one keri database, name allows
                differentiating each instance by name
            temp is boolean, assign to .temp
                True then open in temporary directory, clear on close
                Othewise then open persistent directory, do not clear on close
            headDirPath is optional str head directory pathname for main database
                If not provided use default .HeadDirpath
            mode is int numeric os dir permissions for database directory
            reopen is boolean, IF True then database will be reopened by this init
        """

        super(Pocketer, self).__init__(headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """
        Open sub databases
        """
        super(Pocketer, self).reopen(**kwa)

        self.sers = self.env.open_db(key=b'sers.')
        self.seals = self.env.open_db(key=b'seals.')
        self.sigs = self.env.open_db(key=b'sigs.', dupsort=True)
        self.issus = self.env.open_db(key=b'issus.', dupsort=True)
        self.subjs = self.env.open_db(key=b'subjs.', dupsort=True)
        self.schms = self.env.open_db(key=b'schms.', dupsort=True)

        return self.env

    def getSigs(self, key):
        """
        Use dgKey()
        Return list of indexed witness signatures at key
        Returns empty list if no entry at key
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getVals(self.sigs, key)

    def getSigsIter(self, key):
        """
        Use dgKey()
        Return iterator of indexed witness signatures at key
        Raises StopIteration Error when empty
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getValsIter(self.sigs, key)

    def putSigs(self, key, vals):
        """
        Use dgKey()
        Write each entry from list of bytes indexed witness signatures vals to key
        Adds to existing signatures at key if any
        Returns True If no error
        Apparently always returns True (is this how .put works with dupsort=True)
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.putVals(self.sigs, key, vals)

    def addSig(self, key, val):
        """
        Use dgKey()
        Add indexed witness signature val bytes as dup to key in db
        Adds to existing values at key if any
        Returns True if written else False if dup val already exists
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.addVal(self.sigs, key, val)

    def cntSigs(self, key):
        """
        Use dgKey()
        Return count of indexed witness signatures at key
        Returns zero if no entry at key
        """
        return self.cntVals(self.sigs, key)

    def delSigs(self, key, val=b''):
        """
        Use dgKey()
        Deletes all values at key if val = b'' else deletes dup val = val.
        Returns True If key exists in database (or key, val if val not b'') Else False
        """
        return self.delVals(self.sigs, key, val)

    def getIssus(self, key):
        """
        Use dgKey()
        Return list of indexed witness signatures at key
        Returns empty list if no entry at key
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoVals(self.issus, key)

    def getIssusIter(self, key):
        """
        Use dgKey()
        Return iterator of indexed witness signatures at key
        Raises StopIteration Error when empty
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoValsIter(self.issus, key)

    def putIssus(self, key, vals):
        """
        Use dgKey()
        Write each entry from list of bytes indexed witness signatures vals to key
        Adds to existing signatures at key if any
        Returns True If no error
        Apparently always returns True (is this how .put works with dupsort=True)
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.putIoVals(self.issus, key, vals)

    def addIssu(self, key, val):
        """
        Use dgKey()
        Add indexed witness signature val bytes as dup to key in db
        Adds to existing values at key if any
        Returns True if written else False if dup val already exists
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.addIoVal(self.issus, key, val)

    def cntIssus(self, key):
        """
        Use dgKey()
        Return count of indexed witness signatures at key
        Returns zero if no entry at key
        """
        return self.cntIoVals(self.issus, key)

    def delIssus(self, key):
        """
        Use dgKey()
        Deletes all values at key if val = b'' else deletes dup val = val.
        Returns True If key exists in database (or key, val if val not b'') Else False
        """
        return self.delIoVals(self.issus, key)

    def getSubjs(self, key):
        """
        Use dgKey()
        Return list of indexed witness signatures at key
        Returns empty list if no entry at key
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoVals(self.subjs, key)

    def getSubjsIter(self, key):
        """
        Use dgKey()
        Return iterator of indexed witness signatures at key
        Raises StopIteration Error when empty
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoValsIter(self.subjs, key)

    def putSubjs(self, key, vals):
        """
        Use dgKey()
        Write each entry from list of bytes indexed witness signatures vals to key
        Adds to existing signatures at key if any
        Returns True If no error
        Apparently always returns True (is this how .put works with dupsort=True)
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.putIoVals(self.subjs, key, vals)

    def addSubj(self, key, val):
        """
        Use dgKey()
        Add indexed witness signature val bytes as dup to key in db
        Adds to existing values at key if any
        Returns True if written else False if dup val already exists
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.addIoVal(self.subjs, key, val)

    def cntSubjs(self, key):
        """
        Use dgKey()
        Return count of indexed witness signatures at key
        Returns zero if no entry at key
        """
        return self.cntIoVals(self.subjs, key)

    def delSubjs(self, key):
        """
        Use dgKey()
        Deletes all values at key if val = b'' else deletes dup val = val.
        Returns True If key exists in database (or key, val if val not b'') Else False
        """
        return self.delIoVals(self.subjs, key)

    def getSchms(self, key):
        """
        Use dgKey()
        Return list of indexed witness signatures at key
        Returns empty list if no entry at key
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoVals(self.schms, key)

    def getSchmsIter(self, key):
        """
        Use dgKey()
        Return iterator of indexed witness signatures at key
        Raises StopIteration Error when empty
        Duplicates are retrieved in lexocographic order not insertion order.
        """
        return self.getIoValsIter(self.schms, key)

    def putSchms(self, key, vals):
        """
        Use dgKey()
        Write each entry from list of bytes indexed witness signatures vals to key
        Adds to existing signatures at key if any
        Returns True If no error
        Apparently always returns True (is this how .put works with dupsort=True)
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.putIoVals(self.schms, key, vals)

    def addSchm(self, key, val):
        """
        Use dgKey()
        Add indexed witness signature val bytes as dup to key in db
        Adds to existing values at key if any
        Returns True if written else False if dup val already exists
        Duplicates are inserted in lexocographic order not insertion order.
        """
        return self.addIoVal(self.schms, key, val)

    def cntSchms(self, key):
        """
        Use dgKey()
        Return count of indexed witness signatures at key
        Returns zero if no entry at key
        """
        return self.cntIoVals(self.schms, key)

    def delSchms(self, key):
        """
        Use dgKey()
        Deletes all values at key if val = b'' else deletes dup val = val.
        Returns True If key exists in database (or key, val if val not b'') Else False
        """
        return self.delIoVals(self.schms, key)

    def putSers(self, key, val):
        """
        Use dgKey()
        Write serialized event bytes val to key
        Does not overwrite existing val if any
        Returns True If val successfully written Else False
        Return False if key already exists
        """
        return self.putVal(self.sers, key, val)

    def setSers(self, key, val):
        """
        Use dgKey()
        Write serialized event bytes val to key
        Overwrites existing val if any
        Returns True If val successfully written Else False
        """
        return self.setVal(self.sers, key, val)

    def getSers(self, key):
        """
        Use dgKey()
        Return event at key
        Returns None if no entry at key
        """
        return self.getVal(self.sers, key)

    def delSers(self, key):
        """
        Use dgKey()
        Deletes value at key.
        Returns True If key exists in database Else False
        """
        return self.delVal(self.sers, key)

    def putSeals(self, key, val):
        """
        Use dgKey()
        Write sealialized event bytes val to key
        Does not overwrite existing val if any
        Returns True If val successfully written Else False
        Return False if key already exists
        """
        return self.putVal(self.seals, key, val)

    def setSeals(self, key, val):
        """
        Use dgKey()
        Write sealialized event bytes val to key
        Overwrites existing val if any
        Returns True If val successfully written Else False
        """
        return self.setVal(self.seals, key, val)

    def getSeals(self, key):
        """
        Use dgKey()
        Return event at key
        Returns None if no entry at key
        """
        return self.getVal(self.seals, key)

    def delSeals(self, key):
        """
        Use dgKey()
        Deletes value at key.
        Returns True If key exists in database Else False
        """
        return self.delVal(self.seals, key)


class Wallet:
    """
    Wallet represents all credentials received or verified


    """

    def __init__(self, hab, db: Pocketer = None, name="test", temp=False):
        """
        Create a Wallet associated with a Habitat

        Parameters:
            hab (Habitat) is the local environment associate with this wallet

        """
        self.hab = hab
        self.name = name
        self.temp = temp

        self.db = db if db is not None else Pocketer(name=self.name, temp=self.temp)

    def processCredential(self, creder, prefixer, seqner, diger, sigers):
        """
        Verify the data of the credential against the schema, the SAID of the credential and
        the CESR Proof on the credential and if valid, store the credential

        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential
        """
        if not self.hab.verify(creder, prefixer, seqner, diger, sigers):
            raise UnverifiedProofError("invalid signatures on credential")

        self.saveCredential(creder, prefixer, seqner, diger, sigers)

    def saveCredential(self, creder, prefixer, seqner, diger, sigers):
        """
        Write the credential and associated indicies to the database
        
        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential
        """
        said = creder.said.encode("utf-8")
        schema = creder.schema.encode("utf-8")
        issuer = creder.issuer.encode("utf-8")
        subject = creder.subject["i"].encode("utf-8")
        raw = creder.raw
        self.db.putSers(key=said, val=raw)

        # Signer KEL Location and signatures
        triple = prefixer.qb64b + seqner.qb64b + diger.qb64b
        self.db.putSeals(said, triple)
        self.db.putSigs(said, [siger.qb64b for siger in sigers])  # idempotent

        # Look up indicies
        self.db.addIssu(key=issuer, val=said)
        self.db.addSubj(key=subject, val=said)
        self.db.addSchm(key=schema, val=said)

        logger.info("Credential: %s, Schema: %s,  Saved", creder.said, creder.schema)
        logger.info(creder.pretty())
        print()

    def getCredentials(self, schema=None):
        """
        Return list of (creder, prefixer, seqner, diger, sigers) for each credential
        that matches schema

        Parameters:
            schema: qb64 SAID of the schema for the credential

        """
        saids = self.db.getSchms(key=schema.encode("utf-8"))

        creds = []
        for said in saids:
            raw = self.db.getSers(key=bytearray(said))
            creder = Credentialer(raw=bytes(raw))

            trip = bytearray(self.db.getSeals(bytearray(said)))

            prefixer = Prefixer(qb64b=trip, strip=True)
            seqner = Seqner(qb64b=trip, strip=True)
            diger = Diger(qb64b=trip, strip=True)

            sigs = self.db.getSigs(bytearray(said))
            sigers = [Siger(qb64b=bytearray(sig)) for sig in sigs]

            creds.append((creder, prefixer, seqner, diger, sigers))

        return creds



# -*- encoding: utf-8 -*-
"""
KERI
keri.db.koming module

"""
import json
from dataclasses import dataclass, asdict
from typing import Type, Union, Iterable

import cbor2
import msgpack

from . import dbing
from .. import help
from ..core import coring
from ..help import helping

logger = help.ogler.getLogger()



class KomerBase():
    """
    KomerBase is a base class for Komer (Keyspace Object Mapper) subclasses that
    each use a dataclass as the object mapped via serialization to an dber LMDB
    database subclass.
    Each Komer .schema is a dataclass class reference that is used to define
    the fields in each database entry. The base class is not meant to be instantiated.
    Use an instance of one of the subclasses instead.
    """
    Sep = '.'  # separator for combining key iterables

    def __init__(self, db: dbing.LMDBer, *,
                 subkey: str = 'docs.',
                 schema: Type[dataclass],  # class not instance
                 kind: str = coring.Serials.json,
                 dupsort: bool = False,
                 sep: str = None,
                 **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
        """
        super(KomerBase, self).__init__(**kwa)  # Mixin for Multi-inheritance MRO
        self.db = db
        self.sdb = self.db.env.open_db(key=subkey.encode("utf-8"), dupsort=dupsort)
        self.schema = schema
        self.kind = kind
        self.serializer = self._serializer(kind)
        self.deserializer = self._deserializer(kind)
        self.sep = sep if sep is not None else self.Sep


    def _tokey(self, keys: Union[str, bytes, Iterable]):
        """
        Converts key to key str with proper separators and returns key bytes.
        If key is already str then returns. Else If key is iterable (non-str)
        of strs then joins with separator converts to bytes and returns

        Parameters:
           keys (Union[str, bytes, Iterable]): str, bytes, or Iterable of str.

        """
        if hasattr(keys, "encode"):  # str
            return keys.encode("utf-8")
        elif hasattr(keys, "decode"): # bytes
            return keys
        return (self.sep.join(keys).encode("utf-8"))  # iterable


    def _tokeys(self, key: Union[str, bytes]):
        """
        Converts key bytes to keys tuple of strs by decoding and then splitting
        at separator.

        Returns:
           keys (iterable): of str

        Parameters:
           key (Union[str, bytes]): str or bytes.

        """
        return tuple(key.decode("utf-8").split(self.sep))


    def _serializer(self, kind):
        """
        Parameters:
            kind (str): serialization
        """
        if kind == coring.Serials.mgpk:
            return self.__serializeMGPK
        elif kind == coring.Serials.cbor:
            return self.__serializeCBOR
        else:
            return self.__serializeJSON


    def _deserializer(self, kind):
        """
        Parameters:
            kind (str): deserialization
        """
        if kind == coring.Serials.mgpk:
            return self.__deserializeMGPK
        elif kind == coring.Serials.cbor:
            return self.__deserializeCBOR
        else:
            return self.__deserializeJSON


    def __deserializeJSON(self, val):
        if val is not None:
            val = helping.datify(self.schema, json.loads(bytes(val).decode("utf-8")))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __deserializeMGPK(self, val):
        if val is not None:
            val = helping.datify(self.schema, msgpack.loads(bytes(val)))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __deserializeCBOR(self, val):
        if val is not None:
            val = helping.datify(self.schema, cbor2.loads(bytes(val)))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __serializeJSON(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = json.dumps(helping.dictify(val),
                          separators=(",", ":"),
                          ensure_ascii=False).encode("utf-8")
        return val


    def __serializeMGPK(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = msgpack.dumps(helping.dictify(val))
        return val


    def __serializeCBOR(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = cbor2.dumps(helping.dictify(val))
        return val



class Komer(KomerBase):
    """
    Keyspace Object Mapper factory class.
    """
    def __init__(self,
                 db: dbing.LMDBer, *,
                 subkey: str = 'docs.',
                 schema: Type[dataclass],  # class not instance
                 kind: str = coring.Serials.json,
                 **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
        """
        super(Komer, self).__init__(db=db, subkey=subkey, schema=schema,
                                    kind=kind, dupsort=False, **kwa)


    def put(self, keys: Union[str, Iterable], val: dataclass):
        """
        Puts val at key made from keys. Does not overwrite

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of dataclass of type self.schema as value

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        return (self.db.putVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))


    def pin(self, keys: Union[str, Iterable], val: dataclass):
        """
        Pins (sets) val at key made from keys. Overwrites.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of dataclass of type self.schema as value

        Returns:
            result (bool): True If successful. False otherwise.
        """
        return (self.db.setVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets val at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            val (dataclass):
            None if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (val := mydb.get(keys)) is None:
                raise ExceptionHere
            use val here
        """
        return (self.deserializer(self.db.getVal(db=self.sdb,
                                key=self._tokey(keys))))


    def rem(self, keys: Union[str, Iterable]):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return (self.db.delVal(db=self.sdb, key=self._tokey(keys)))


    def getItemIter(self):
        """
        Return iterator over the all the items in subdb

        Returns:
            iterator: of tuples of keys tuple and val dataclass instance for
            each entry in db. Raises StopIteration when done

        Example:
            if key in database is "a.b" and val is serialization of dataclass
               with attributes x and y then returns
               (("a","b"), dataclass(x=1,y=2))
        """
        for key, val in self.db.getAllItemIter(db=self.sdb, split=False):
            yield (self._tokeys(key), self.deserializer(val))




class DupKomer(KomerBase):
    """
    Duplicate Keyspace Object Mapper factory class that supports multiple entries
    a given database key (lmdb dupsort == True).
    """
    def __init__(self,
             db: dbing.LMDBer, *,
             subkey: str = 'recs.',
             schema: Type[dataclass],  # class not instance
             kind: str = coring.Serials.json,
             **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
        """
        super(DupKomer, self).__init__(db=db, subkey=subkey, schema=schema,
                                       kind=kind, dupsort=True, **kwa)



    def put(self, keys: Union[str, Iterable], vals: list):
        """
        Puts all vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        Apparently always returns True (how .put works with dupsort=True)

        """
        vals = [self.serializer(val) for val in vals]
        return (self.db.putVals(db=self.sdb,
                                key=self._tokey(keys),
                                vals=vals))


    def add(self, keys: Union[str, Iterable], val: dataclass):
        """
        Add val to vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of type self.schema

        Returns:
            result (bool): True means unique value among duplications,
                              False means duplicte of same value already exists.

        """
        return (self.db.addVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))


    def pin(self, keys: Union[str, Iterable], vals: list):
        """
        Pins (sets) vals at key made from keys. Overwrites. Removes all
        pre-existing dup vals and replaces them with vals

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        """
        key = self._tokey(keys)
        self.db.delVals(db=self.sdb, key=key)  # delete all values
        vals = [self.serializer(val) for val in vals]
        return (self.db.putVals(db=self.sdb,
                                key=key,
                                vals=vals))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets dup vals list at key made from keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            vals (list):  each item in list is instance of type self.schema
                          empty list if no entry at keys

        """
        vals = self.db.getVals(db=self.sdb, key=self._tokey(keys))
        vals = [self.deserializer(val) for val in vals]
        return vals


    def getIter(self, keys: Union[str, Iterable]):
        """
        Gets dup vals iterator at key made from keys

        Duplicates are retrieved in lexocographic order not insertion order.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            iterator:  vals each of type self.schema. Raises StopIteration when done

        """
        for val in self.db.getValsIter(db=self.sdb, key=self._tokey(keys)):
            yield self.deserializer(val)


    def cnt(self, keys: Union[str, Iterable]):
        """
        Return count of dup values at key made from keys, zero otherwise

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
        """
        return (self.db.cntVals(db=self.sdb, key=self._tokey(keys)))


    def rem(self, keys: Union[str, Iterable], val=None):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass):  instance of dup val at key to delete
                              if val is None then remove all values at key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise

        """
        if val is not None:
            val = self.serializer(val)
        else:
            val = b''
        return (self.db.delVals(db=self.sdb, key=self._tokey(keys), val=val))


    def getItemIter(self):
        """
        Return iterator over the all the items in subdb. Each duplicate at a
        given key is yielded as a separate item.

        Returns:
            iterator: of tuples of keys tuple and val dataclass instance for
            each entry in db. Raises StopIteration when done

        Example:
            if key in database is "a.b" and val is serialization of dataclass
               with attributes x and y then returns
               (("a","b"), dataclass(x=1,y=2))
        """
        for key, val in self.db.getAllItemIter(db=self.sdb, split=False):
            yield (self._tokeys(key), self.deserializer(val))





# -*- encoding: utf-8 -*-
"""
KERI
keri.demo.demoing module

Utilities for demos
"""
import os
import argparse
import logging

from keri import __version__
from keri.app import directing
from keri.demo import demoing
from keri import help


def runDemo(name="bob", remote=5621, local=5620, expire=0.0):
    """
    Setup and run one demo controller for Bob
    """
    secrets = [
                'ArwXoACJgOleVZ2PY7kXn7rA0II0mHYDhc6WrBH8fDAc',
                'A6zz7M08-HQSFq92sJ8KJOT2cZ47x7pXFQLPB0pckB3Q',
                'AcwFTk-wgk3ZT2buPRIbK-zxgPx-TKbaegQvPEivN90Y',
                'Alntkt3u6dDgiQxTATr01dy8M72uuaZEf9eTdM-70Gk8',
                'A1-QxDkso9-MR1A8rZz_Naw6fgaAtayda8hrbkRVVu1E',
                'AKuYMe09COczwf2nIoD5AE119n7GLFOVFlNLxZcKuswc',
                'AxFfJTcSuEE11FINfXMqWttkZGnUZ8KaREhrnyAXTsjw',
                'ALq-w1UKkdrppwZzGTtz4PWYEeWm0-sDHzOv5sq96xJY'
                ]

    doers = demoing.setupDemoController(secrets=secrets,
                                        name=name,
                                        remotePort=remote,
                                        localPort=local)

    directing.runController(doers=doers, expire=expire)



def parseArgs(version=__version__):
    d = "Runs KERI direct mode demo controller.\n"
    d += "Example:\nkeri_bob -r 5621 -l 5620 --e 10.0'\n"
    p = argparse.ArgumentParser(description=d)
    p.add_argument('-V', '--version',
                   action='version',
                   version=version,
                   help="Prints out version of script runner.")
    p.add_argument('-r', '--remote',
                   action='store',
                   default=5621,
                   help="Remote port number the client connects to. Default is 5621.")
    p.add_argument('-l', '--local',
                   action='store',
                   default=5620,
                   help="Local port number the server listens on. Default is 5620.")
    p.add_argument('-e', '--expire',
                   action='store',
                   default=0.0,
                   help="Expire time for demo. 0.0 means not expire. Default is 0.0.")
    p.add_argument('-n', '--name',
                   action='store',
                   default="bob",
                   help="Name of controller. Default is bob. Choices are bob, sam, or eve.")


    args = p.parse_args()

    return args


def main():
    args = parseArgs(version=__version__)

    help.ogler.level = logging.INFO
    help.ogler.reopen(name=args.name, temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Demo for %s listening on %s connecting to "
                 "%s.******\n\n", args.name, args.local, args.remote)

    runDemo(name=args.name,
            remote=args.remote,
            local=args.local,
            expire=args.expire)

    logger.info("\n******* Ended Demo for %s listening on %s connecting to "
                 "%s.******\n\n", args.name, args.local, args.remote)


if __name__ == "__main__":
    main()




# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""
import argparse
import logging

from keri import __version__
from keri import help
from keri.app import directing, indirecting

d = "Runs KERI witness controller.\n"
d += "Example:\nwitness -H 5631 -t 5632\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument('-V', '--version',
                    action='version',
                    version=__version__,
                    help="Prints out version of script runner.")
parser.add_argument('-H', '--http',
                    action='store',
                    default=5631,
                    help="Local port number the HTTP server listens on. Default is 5631.")
parser.add_argument('-T', '--tcp',
                    action='store',
                    default=5632,
                    help="Local port number the HTTP server listens on. Default is 5632.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="witness",
                    help="Name of controller. Default is witness.")


def launch(args):
    help.ogler.level = logging.INFO
    help.ogler.reopen(name=args.name, temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Witness for %s listening: http/%s, tcp/%s "
                ".******\n\n", args.name, args.http, args.tcp)

    runWitness(name=args.name,
               tcp=int(args.tcp),
               http=int(args.http))

    logger.info("\n******* Ended Witness for %s listening: http/%s, tcp/%s"
                ".******\n\n", args.name, args.http, args.tcp)


def runWitness(name="witness", tcp=5631, http=5632, expire=0.0):
    """
    Setup and run one witness
    """

    doers = indirecting.setupWitness(name=name,
                                     tcpPort=tcp,
                                     httpPort=http)

    directing.runController(doers=doers, expire=expire)



# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse
import os

parser = argparse.ArgumentParser(description='Print version of KLI')
parser.set_defaults(handler=lambda args: exit())


def exit():
    os._exit(os.EX_OK)



# -*- encoding: utf-8 -*-
"""
keri.kli.common.config

"""
import json
from dataclasses import dataclass, asdict
from typing import Optional, List


@dataclass
class Config:
    """
    Config representation of prefix configuration
    """
    name: str  # Alias for prefix
    witnessExpire: Optional[int] = 60  # timeout for witness connection
    witnesses: Optional[List[str]] = None  # list of witness nodes
    transferable: Optional[bool] = None  # Use transferable derivation code

    def __iter__(self):
        return iter(asdict(self))


def loadConfig(file: str) -> Config:
    """
        Parameters:
            file (str): file location to load from
    """
    cfg = None
    if file != '':
        with open(file, "r") as f:
            cfg = Config(**json.loads(f.read()))
    return cfg



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import shutil

from keri.app import keeping
from keri.db import basing

parser = argparse.ArgumentParser(description='Delete existing KERI database')
parser.set_defaults(handler=lambda args: purge(args.name))
parser.add_argument('--name', '-n', help='Humane reference')


def purge(name):
    with basing.openDB(name=name, temp=False) as db, keeping.openKS(name=name, temp=False) as ks:
        db.close(clear=True)
        ks.close(clear=True)

    try:
        shutil.rmtree(f"/usr/local/var/keri/db/{name}_tel")
    except FileNotFoundError:
        pass





# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio.base import doing

from keri import kering
from keri.db import basing
from ... import habbing, keeping, agenting, indirecting, directing

parser = argparse.ArgumentParser(description='Rotate keys')
parser.set_defaults(handler=lambda args: handler(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--proto', '-p', help='Protocol to use when propagating ICP to witnesses [tcp|http] (defaults '
                                          'http)', default="tcp")
parser.add_argument('--erase', '-e', help='if this option is provided stale keys will be erased', default=False)
parser.add_argument('--count', '-C', help='', default=None, type=int, required=False)
parser.add_argument('--sith', '-s', help='', default=None, type=int, required=False)
parser.add_argument('--toad', '-t', help='', default=None, type=int, required=False)
parser.add_argument('--witnesses', '-w', help='New set of witnesses, replaces all existing witnesses.  Can appear '
                                              'multiple times', metavar="<prefix>", default=[],
                    action="append", required=False)
parser.add_argument('--witness-cut', '-c', help='Witnesses to remove.  Can appear multiple times', metavar="<prefix>",
                    default=[],
                    action="append", required=False)
parser.add_argument('--witness-add', '-a', help='Witnesses to add.  Can appear multiple times', metavar="<prefix>",
                    default=[],
                    action="append", required=False)
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=[], action="store", required=False)


def handler(args):
    name = args.name

    rotDoer = RotateDoer(name=name, proto=args.proto, wits=args.witnesses, cuts=args.witness_cut, adds=args.witness_add,
                         sith=args.sith, count=args.count, toad=args.toad, erase=args.erase)

    doers = [rotDoer]
    directing.runController(doers=doers, expire=0.0)


class RotateDoer(doing.DoDoer):

    def __init__(self, name, proto, sith=None, count=None, erase=None,
                 toad=None, wits=None, cuts=None, adds=None, data=None, tock=0.0, **kwa):
        """

        Parameters:
            name is human readable str of identifier
            proto is tcp or http method for communicating with Witness
            sith is next signing threshold as int or str hex or list of str weights
            count is int next number of signing keys
            erase is Boolean True means erase stale keys
            toad is int or str hex of witness threshold after cuts and adds
            cuts is list of qb64 pre of witnesses to be removed from witness list
            adds is list of qb64 pre of witnesses to be added to witness list
            data is list of dicts of committed data such as seals
       """

        self.name = name
        self.proto = proto
        self.sith = sith
        self.count = count
        self.erase = erase
        self.toad = toad
        self.data = data

        self.wits = wits if wits is not None else []
        self.cuts = cuts if cuts is not None else []
        self.adds = adds if adds is not None else []

        ks = keeping.Keeper(name=self.name, temp=False)  # not opened by default, doer opens
        self.ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
        db = basing.Baser(name=self.name, temp=False, reload=True)  # not opened by default, doer opens
        self.dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

        self.hab = habbing.Habitat(name=self.name, ks=ks, db=db, temp=False, create=False)
        self.habDoer = habbing.HabitatDoer(habitat=self.hab)  # setup doer
        doers = [self.ksDoer, self.dbDoer, self.habDoer, doing.doify(self.rotateDo)]

        super(RotateDoer, self).__init__(doers=doers, **kwa)


    def rotateDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.wits:
            if self.adds or self.cuts:
                raise kering.ConfigurationError("you can only specify witnesses or cuts and add")
            ewits = self.hab.kever.lastEst.wits

            # wits= [a,b,c]  wits=[b, z]
            self.cuts = set(self.wits) & set(ewits)
            self.adds = set(self.wits) - set(ewits)

        msg = self.hab.rotate(sith=self.sith, count=self.count, erase=self.erase, toad=self.toad,
                              cuts=list(self.cuts), adds=list(self.adds), data=self.data)

        if self.proto == "tcp":
            mbx = None
            witDoer = agenting.WitnessReceiptor(hab=self.hab, klas=agenting.TCPWitnesser, msg=msg)
            self.extend(doers=[witDoer])
            yield self.tock
        else:  # "http"
            mbx = indirecting.MailboxDirector(hab=self.hab)
            witDoer = agenting.WitnessReceiptor(hab=self.hab, klas=agenting.TCPWitnesser, msg=msg)
            self.extend(doers=[mbx, witDoer])
            yield self.tock

        while not witDoer.done:
            _ = yield self.tock


        print(f'Prefix  {self.hab.pre}')
        for idx, verfer in enumerate(self.hab.kever.verfers):
            print(f'\tPublic key {idx+1}:  {verfer.qb64}')

        toRemove = [self.ksDoer, self.dbDoer, self.habDoer, witDoer]
        if mbx:
            toRemove.append(mbx)

        self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import logging

from hio.base import doing
from hio.core.tcp import clienting

from keri.app import habbing
from keri.app.cli.commands.init import KLIRecord
from keri.db import koming

parser = argparse.ArgumentParser(description='view info')
parser.set_defaults(handler=lambda args: handler())

logging.basicConfig(
    filename='klid.log',
    level=logging.DEBUG, format='%(asctime)s [%(levelname)s] %(message)s',
)


def handler():
    infoIst = InfoIst(tock=0.03125)
    infoIst.do()

    return


class InfoIst(doing.Doist):

    def __init__(self, real=False, limit=None, doers=None, **kwa):
        # add config
        self.hab = habbing.Habitat(name='kli', temp=False)
        kli = koming.Komer(db=self.hab.db, schema=KLIRecord, subkey='kli.').get((self.hab.pre,))

        print(kli.host, kli.port)

        super().__init__(real, limit, doers, **kwa)

        self.client = clienting.Client(tymth=self.tymen(), host='127.0.0.1', port=5678)
        clientDoer = clienting.ClientDoer(client=self.client)

        self.extend([clientDoer, doing.doify(self.infoDoer)])


    def infoDoer(self, tymth=None, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        while not self.client.connected:
            (yield self.tock)

        msg = dict(
            cmd='info',
            args=dict(),
        )

        self.hab.endorse(msg)

        self.client.tx(msg)

        self.client.close()

        return



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.directing module

simple direct mode demo support classes
"""

import json
import os
from urllib import parse

from hio.core import wiring
from hio.core.serial import serialing
from hio.core.tcp import clienting, serving

from .. import help
from ..app import habbing, keeping, directing
from ..core import coring, scheming
from ..core.scheming import CacheResolver
from ..db import basing
from ..help import helping
from ..peer import exchanging
from ..vc import proving, handling, walleting
from ..vc.handling import RequestHandler

logger = help.ogler.getLogger()


def setupDemoController(secrets, name="who", remotePort=5621, localPort=5620,
                        indirect=False, remotePre=""):
    """
    Setup and return doers list to run controller
    """
    secrecies = []
    for secret in secrets:  # convert secrets to secrecies
        secrecies.append([secret])

    # setup databases for dependency injection
    ks = keeping.Keeper(name=name, temp=True)  # not opened by default, doer opens
    ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
    db = basing.Baser(name=name, temp=True)  # not opened by default, doer opens
    dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

    # setup habitat
    hab = habbing.Habitat(name=name, ks=ks, db=db, temp=True, secrecies=secrecies)
    habDoer = habbing.HabitatDoer(habitat=hab)  # setup doer

    # setup wirelog to create test vectors
    path = os.path.dirname(__file__)
    path = os.path.join(path, 'logs')

    wl = wiring.WireLog(samed=True, filed=True, name=name, prefix='demo', reopen=True,
                        headDirPath=path)
    wireDoer = wiring.WireLogDoer(wl=wl)

    client = clienting.Client(host='127.0.0.1', port=remotePort, wl=wl)
    clientDoer = clienting.ClientDoer(client=client)

    if name == 'bob':
        director = BobDirector(hab=hab, client=client, tock=0.125)
    elif name == "sam":
        director = SamDirector(hab=hab, client=client, tock=0.125)
    elif name == 'eve':
        director = EveDirector(hab=hab, client=client, tock=0.125)
    else:
        raise ValueError("Invalid director name={}.".format(name))

    reactor = directing.Reactor(hab=hab, client=client, indirect=indirect)

    server = serving.Server(host="", port=localPort, wl=wl)
    serverDoer = serving.ServerDoer(server=server)
    directant = directing.Directant(hab=hab, server=server)
    # Reactants created on demand by directant

    logger.info("\nDirect Mode demo of %s:\nNamed %s on TCP port %s to port %s.\n\n",
                hab.pre, hab.name, localPort, remotePort)

    return [ksDoer, dbDoer, habDoer, wireDoer, clientDoer, director, reactor,
            serverDoer, directant]


class BobDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist dependencies
            self.tock = tock
            # tyme = self.tyme

            # recur context
            tyme = (yield (self.tock))  # yields tock then waits for next send

            logger.info("%s:\nWaiting for connection to remote  %s.\n\n", self.hab.pre, self.client.ha)
            while (not self.client.connected):
                tyme = (yield (self.tock))

            logger.info("%s:\nConnected to %s.\n\n", self.hab.pre, self.client.ha)

            self.sendOwnInception()  # Inception Event
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.interact()  # Interaction event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            # create a bunch of out of order messages to test out of order escrow
            msgs = []
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.interact())  # Interaction event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.interact())  # Interaction event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.rotate())  # Rotation event
            msgs.append(self.hab.interact())

            msgs.reverse()  # reverse the order

            for msg in msgs:
                self.client.tx(msg)  # send to connected remote
                logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
                tyme = (yield (self.tock))

            tyme = (yield (self.tock))

        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration


class SamDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock
            # tyme = self.tyme

            # recur context
            tyme = (yield (self.tock))  # yields tock then waits

            while (not self.client.connected):
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield (self.tock))

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)

            self.sendOwnInception()  # Inception Event
            tyme = (yield (self.tock))

            msg = self.hab.interact()  # Interaction Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.interact()  # Interaction Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration


class IanDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def __init__(self, recipientIdentifier, lei, hab, issuer, witnessClient, peerClient, **kwa):
        super().__init__(hab, witnessClient, **kwa)
        self.peerClient = peerClient
        self.recipientIdentifier = recipientIdentifier
        self.issuer = issuer
        self.lei = lei
        if self.tymth:
            self.peerClient.wind(self.tymth)

    def wind(self, tymth):
        """
        Inject new tymist.tymth as new ._tymth. Changes tymist.tyme base.
        Updates winds .tymer .tymth
        """
        super(IanDirector, self).wind(tymth)
        self.peerClient.wind(tymth)

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock

            # recur context
            tyme = (yield self.tock)  # yields tock then waits

            while not self.client.connected:
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield self.tock)

            while not self.peerClient.connected:
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.peerClient.ha)
                tyme = (yield self.tock)

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)

            self.sendOwnInception()  # Inception Event
            tyme = (yield self.tock)

            msg = self.issuer.ianchor
            # send to connected remote
            self.client.tx(msg)
            logger.info("%s: %s sent event:\n%s\n\n", self.hab.name, self.hab.pre, bytes(msg))
            tyme = (yield self.tock)

            tevt = self.issuer.incept
            self.client.tx(tevt)
            logger.info("%s: %s sent event:\n%s\n\n", self.hab.name, self.hab.pre, bytes(tevt))
            tyme = (yield self.tock)

            msg = self.hab.interact()  # Interaction Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield self.tock)

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield self.tock)

            msg = self.hab.rotate()  # Rotation Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield self.tock)

            now = helping.nowIso8601()
            jsonSchema = scheming.JSONSchema(resolver=scheming.jsonSchemaCache)
            ref = scheming.jsonSchemaCache.resolve("EeCCZi1R5xHUlhsyQNm_7NrUQTEKZH5P9vBomnc9AihY")
            schemer = scheming.Schemer(raw=ref)

            # Build the credential subject and then the Credentialer for the full credential
            credSubject = dict(
                id=self.recipientIdentifier,  # this needs to be generated from a KEL
                lei=self.lei
            )

            creder = proving.credential(issuer=self.hab.pre,
                                        schema=schemer.said,
                                        subject=credSubject,
                                        typ=jsonSchema)

            msg = self.hab.endorse(serder=creder)

            tevt, kevt = self.issuer.issue(vcdig=creder.said)
            self.client.tx(kevt)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(kevt))
            tyme = (yield self.tock)

            self.client.tx(tevt)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(tevt))
            tyme = (yield self.tock)

            pl = dict(
                vc=[handling.envelope(msg, typ=jsonSchema)]
            )


            cloner = self.hab.db.clonePreIter(pre=self.hab.pre, fn=0)  # create iterator at 0
            msgs = bytearray()  # outgoing messages
            for msg in cloner:
                msgs.extend(msg)

            # send to connected peer remote
            self.peerClient.tx(msgs)
            logger.info("%s: %s sent event:\n%s\n\n", self.hab.name, self.hab.pre, bytes(msgs))
            tyme = (yield self.tock)

            excSrdr = exchanging.exchange(route="/credential/issue", payload=pl)
            excMsg = self.hab.sanction(excSrdr)

            self.peerClient.tx(excMsg)
            logger.info("%s: %s sent event:\n%s\n\n", self.hab.name, self.hab.pre, bytes(excMsg))
            tyme = (yield self.tock)

            logger.info("%s:\n\n\n Sent Verifiable Credential for LEI: %s to %s.\n\n",
                        self.hab.pre, self.lei, self.recipientIdentifier)

            console = serialing.Console()
            console.reopen()
            while console.get().decode('utf-8') != "r":
                (yield self.tock)

            tevt, kevt = self.issuer.revoke(vcdig=creder.said)
            logger.info("%s:\n\n\n Revoked Verifiable Credential for LEI: %s.\n\n",
                        self.hab.pre, self.lei)
            (yield self.tock)

            self.client.tx(kevt)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(kevt))
            (yield self.tock)

            self.client.tx(tevt)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(tevt))
            (yield self.tock)

        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration


class HanDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    issuerpre = "ExwBAYqvPpaPpGmBCixIiC_xpcDto8YUxLoNJgE2FOKo"

    def __init__(self, wallet, hab, client, exchanger, **kwa):
        super().__init__(hab, client, **kwa)
        self.wallet = wallet
        self.exchanger = exchanger

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock
            tyme = (yield self.tock)  # yields tock then waits

            while not self.client.connected:
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield self.tock)

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)
            print(f'{self.hab.name}\'s Wallet ({self.hab.pre}) ')
            print()

            msg = self.hab.query(self.issuerpre, res="logs")  # Query for remote pre Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            (yield self.tock)


        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration


class VicDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def __init__(self, hab, witnessClient, peerClient, verifier, exchanger, jsonSchema, proofs, **kwa):
        """

            verifier is Verifier instance of local controller's TEL context
        """
        super().__init__(hab, witnessClient, **kwa)
        self.peerClient = peerClient
        self.verifier = verifier
        self.exchanger = exchanger
        self.jsonSchema = jsonSchema
        self.proofs = proofs
        if self.tymth:
            self.peerClient.wind(self.tymth)


    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock

            tyme = (yield self.tock)  # yields tock then waits

            while not self.client.connected:
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield self.tock)

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)

            while not self.peerClient.connected:
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.peerClient.ha)
                tyme = (yield self.tock)

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.peerClient.ha)

            msg = self.hab.makeOwnEvent(sn=0)
            # send to connected remote
            self.peerClient.tx(msg)
            logger.info("%s: %s sent event:\n%s\n\n", self.hab.name, self.hab.pre, bytes(msg))
            tyme = (yield self.tock)

            ref = scheming.jsonSchemaCache.resolve("EeCCZi1R5xHUlhsyQNm_7NrUQTEKZH5P9vBomnc9AihY")
            schemer = scheming.Schemer(raw=ref)

            pl = dict(
                input_descriptors=[
                    dict(x=schemer.said)
                ]
            )

            logger.info("%s: \n requesting presentation for schema %s\n\n", self.hab.pre, schemer.said)

            excSrdr = exchanging.exchange(route="/presentation/request", payload=pl)
            excMsg = self.hab.sanction(excSrdr)

            self.peerClient.tx(excMsg)
            tyme = (yield self.tock)

            while not self.proofs:
                logger.info("%s:\n waiting for proof presentation from %s.\n\n",
                            self.hab.pre, self.peerClient.ha)
                tyme = (yield self.tock)

            _, presentation = self.proofs.pop()

            vc = presentation["vc"]
            body = vc["d"]
            proof = bytearray(presentation["proof"].encode("utf-8"))

            creder = proving.Credentialer(crd=vc, typ=self.jsonSchema)
            prefixer, seqner, diger, isigers = walleting.parseProof(proof)

            vcid = creder.said

            issuerPre = creder.issuer
            issuerpre = issuerPre.removeprefix("did:keri:")

            regk = creder.status["id"]

            msg = self.hab.query(issuerpre, res="logs")  # Query for remote pre Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            logger.info("Loading Registry and Credential TEL")
            msg = self.verifier.query(regk,
                                      vcid,
                                      res="tels")
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))

            while regk not in self.verifier.tevers:
                logger.info("%s:\n waiting for retrieval of TEL %s.\n\n",
                            self.hab.pre, regk)
                tyme = (yield (self.tock))

            valid = self.hab.verify(creder, prefixer, seqner, diger, isigers)
            if valid is True:
                sub = creder.subject
                logger.info("%s:\n\n\n Valid vLEI credential for LEI: %s.\n\n",
                            self.hab.pre, sub["lei"])
            else:
                logger.error("%s:\n\n\n Invalid vLEI credential.\n\n",
                             self.hab.pre)

        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration



class CamDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def __init__(self, remotePre, hab, client, **kwa):
        super().__init__(hab, client, **kwa)
        self.remotePre = remotePre

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock
            # tyme = self.tyme

            # recur context
            tyme = (yield (self.tock))  # yields tock then waits

            while (not self.client.connected):
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield (self.tock))

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)

            msg = self.hab.query(self.remotePre, res="logs")  # Query for remote pre Event
            self.client.tx(msg)  # send to connected remote
            logger.info("%s sent event:\n%s\n\n", self.hab.pre, bytes(msg))
            tyme = (yield (self.tock))



        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration


class EveDirector(directing.Director):
    """
    Direct Mode KERI Director (Contextor, Doer) with TCP Client and Kevery
    Generator logic is to iterate through initiation of events for demo

    Inherited Attributes:
        .hab is Habitat instance of local controller's context
        .client is TCP client instance. Assumes operated by another doer.

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:

    Inherited Methods:
        .__call__ makes instance callable return generator
        .do is generator function returns generator

    Methods:

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
    """

    def do(self, tymth=None, tock=0.0, **opts):
        """
        Generator method to run this doer
        Calling this method returns generator
        """
        try:
            # enter context
            self.wind(tymth)  # change tymist and dependencies
            self.tock = tock
            # tyme = self.tyme

            # recur context after first yield
            tyme = (yield (tock))

            while (not self.client.connected):
                logger.info("%s:\n waiting for connection to remote %s.\n\n",
                            self.hab.pre, self.client.ha)
                tyme = (yield (self.tock))

            logger.info("%s:\n connected to %s.\n\n", self.hab.pre, self.client.ha)
            tyme = (yield (self.tock))

        except GeneratorExit:  # close context, forced exit due to .close
            pass

        except Exception:  # abort context, forced exit due to uncaught exception
            raise

        finally:  # exit context,  unforced exit due to normal exit of try
            pass

        return True  # return value of yield from, or yield ex.value of StopIteration



# -*- encoding: utf-8 -*-
"""
keri.peer.httping module

"""
import json
import random
from dataclasses import dataclass

import falcon
from hio.base import doing
from hio.core import http
from hio.help import helping, Hict, decking

from .. import help
from .. import kering
from ..app import obtaining
from ..core import parsing, coring
from ..core.coring import Ilks
from ..help.helping import nowIso8601
from ..peer import exchanging

logger = help.ogler.getLogger()

CESR_CONTENT_TYPE = "application/cesr+json"
CESR_ATTACHMENT_HEADER = "CESR-ATTACHMENT"
CESR_DATE_HEADER = "CESR-DATE"
CESR_RECIPIENT_HEADER = "CESR-RECIPIENT"


class AgentExnServer(doing.DoDoer):
    """
    Peer 2 Peer HTTP Server that allows for handler registration of `exn` messages by providing an Exchanger
    that is configured with Handlers for all message types to handle.


    """

    RoutePrefix = "/exn"

    def __init__(self, exc, app, rep, **kwa):
        """
        Registers all behaviors in provided Exchanged as routes to be handled in the provided Falcon app.
        POST requests are extracted and mapped to `exn` messages that are passed to the provided Exchanger.

        Parameters:
            exc (Exchanger):   an the instance of Exchanger configured to handle behaviors
            rep (Respondant):  Handles response messages from exchange behaviors
            app (falcon.App):  app to use for route registration

        """
        self.exc = exc
        self.rep = rep
        self.rxbs = bytearray()

        self.app = app if app is not None else falcon.App(cors_enable=True)

        # for route in exc.routes:
        #     self.app.add_route(self.RoutePrefix + route, self)

        self.parser = parsing.Parser(ims=self.rxbs,
                                     framed=True,
                                     kvy=None,
                                     tvy=None,
                                     exc=self.exc)

        self.app.add_sink(prefix=self.RoutePrefix, sink=self.on_post)


        doers = [doing.doify(self.exchangerDo), doing.doify(self.msgDo)]

        super(AgentExnServer, self).__init__(doers=doers, **kwa)


    def msgDo(self, tymth=None, tock=0.0, **opts):
        """
        Returns doifiable Doist compatibile generator method (doer dog) to process
            incoming message stream of .kevery

        Doist Injected Attributes:
            g.tock = tock  # default tock attributes
            g.done = None  # default done state
            g.opts

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters

        Usage:
            add result of doify on this method to doers list
        """
        if self.parser.ims:
            logger.info("Client exn-http received:\n%s\n...\n", self.parser.ims[:1024])
        done = yield from self.parser.parsator()  # process messages continuously
        return done  # should nover get here except forced close


    def exchangerDo(self, tymth=None, tock=0.0, **opts):
        """
         Returns doifiable Doist compatibile generator method (doer dog) to process
            .tevery.cues deque

        Doist Injected Attributes:
            g.tock = tock  # default tock attributes
            g.done = None  # default done state
            g.opts

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters

        Usage:
            add result of doify on this method to doers list
        """
        while True:
            for rep in self.exc.processResponseIter():
                self.rep.msgs.append(rep)
                yield  # throttle just do one cue at a time
            yield


    def on_post(self, req, rep):
        """
        Handles POST requests by generating an `exn` message from the request and passing it to the Exchanger

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        """
        if req.method == 'OPTIONS':
            rep.status = falcon.HTTP_200
            return


        cr = parseCesrHttpRequest(req=req, prefix=self.RoutePrefix)

        serder = exchanging.exchange(route=cr.resource, date=cr.date, payload=cr.payload, recipient=cr.recipient)
        msg = bytearray(serder.raw)
        msg.extend(cr.attachments.encode("utf-8"))

        self.rxbs.extend(msg)

        rep.status = falcon.HTTP_202  # This is the default status


@dataclass
class CesrRequest:
    resource: str
    recipient: str
    date: str
    payload: dict
    modifiers: dict
    attachments: str


def parseCesrHttpRequest(req, prefix=None):
    """
    Parse Falcon HTTP request and create a CESR message from the body of the request and the two
    CESR HTTP headers (Date, Attachment).

    Parameters
        req (falcon.Request) http request object in CESR format:

    """
    if req.content_type != CESR_CONTENT_TYPE:
        raise falcon.HTTPError(falcon.HTTP_NOT_ACCEPTABLE,
                               title="Content type error",
                               description="Unacceptable content type.")

    try:
        data = json.load(req.bounded_stream)
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               title="Malformed JSON",
                               description="Could not decode the request body. The "
                                           "JSON was incorrect.")

    resource = req.path
    if prefix is not None:
        resource = resource.removeprefix(prefix)

    if CESR_DATE_HEADER not in req.headers:
        raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED,
                               title="Date error",
                               description="Missing required date header.")

    dt = req.headers[CESR_DATE_HEADER]

    if CESR_ATTACHMENT_HEADER not in req.headers:
        raise falcon.HTTPError(falcon.HTTP_PRECONDITION_FAILED,
                               title="Attachment error",
                               description="Missing required attachment header.")
    attachment = req.headers[CESR_ATTACHMENT_HEADER]
    recipient = req.headers[CESR_RECIPIENT_HEADER] if CESR_RECIPIENT_HEADER in req.headers else None

    cr = CesrRequest(
        resource=resource,
        recipient=recipient,
        date=dt,
        payload=data,
        modifiers=req.params,
        attachments=attachment)

    return cr


def createCESRRequest(msg, client, date=None):
    """
    Turns a KERI message into a CESR http request against the provided hio http Client

    Parameters
       msg:  KERI message parsable as Serder.raw
       client: hio http Client that will send the message as a CESR request

    """

    dt = date if date is not None else nowIso8601()
    try:
        serder = coring.Serder(raw=msg)
    except kering.ShortageError as ex:  # need more bytes
        raise kering.ExtractionError("unable to extract a valid message to send as HTTP")
    else:  # extracted successfully
        del msg[:serder.size]  # strip off event from front of ims

    ilk = serder.ked["t"]
    attachments = bytearray(msg)
    query = serder.ked["q"] if "q" in serder.ked else None

    if ilk in (Ilks.exn,):
        resource = "/" + ilk + serder.ked['r']
        body = json.dumps(serder.ked["d"]).encode("utf-8")
        dt = serder.ked["dt"]
    elif ilk in (Ilks.req,):
        resource = "/" + ilk + serder.ked['r']
        body = serder.raw
    elif ilk in (Ilks.icp, Ilks.rot, Ilks.ixn, Ilks.dip, Ilks.drt, Ilks.ksn, Ilks.rct):
        resource = "/kel"
        body = serder.raw
    elif ilk in (Ilks.vcp, Ilks.vrt, Ilks.iss, Ilks.rev, Ilks.bis, Ilks.brv):
        resource = "/tel"
        body = serder.raw
    else:
        raise kering.InvalidEventTypeError("Event type {} is not handled by http clients".format(ilk))

    headers = Hict([
        ("Content-Type", CESR_CONTENT_TYPE),
        ("Content-Length", len(body)),
        (CESR_DATE_HEADER, dt),
        (CESR_ATTACHMENT_HEADER, attachments)
    ])

    if "i" in serder.ked:
        headers[CESR_RECIPIENT_HEADER] = serder.pre

    client.request(
        method="POST",
        path=resource,
        quargs=query,
        headers=headers,
        body=body
    )


class MailboxServer(doing.DoDoer):
    """
    Message storage for Witnesses.  Provides an inbox service for storing messages for an identifier.

    """

    def __init__(self, mbx: exchanging.Mailboxer, app=None, **kwa):
        """
        Create Mailbox server for storing messages on a Witness for a witnessed
        identifier.

        Parameters:
             app(falcon.App): REST app to register routes with

        """

        self.mbx = mbx
        self.rxbs = bytearray()

        self.app = app if app is not None else falcon.App(cors_enable=True)

        self.app.add_route("/req/mbx", self)

        doers = []

        super(MailboxServer, self).__init__(doers=doers, **kwa)

    def on_get(self, req, rep):
        """
        Handles GET requests as a stream of SSE events

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        """
        rep.stream = self.mailboxGenerator(query=req.params, resp=rep)

    def on_post(self, req, rep):
        """
        Handles GET requests as a stream of SSE events

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        """
        cr = parseCesrHttpRequest(req=req)

        q = cr.payload['q']

        rep.stream = self.mailboxGenerator(query=q, resp=rep)


    @helping.attributize
    def mailboxGenerator(self, me, query=None, resp=None):
        """

        Parameters:
            me:
            query:
            resp:

        """
        pre = coring.Prefixer(qb64=query["i"])
        idx = int(query["s"]) if "s" in query else 0

        me._status = http.httping.OK

        headers = Hict()
        headers['Content-Type'] = "text/event-stream"
        headers['Cache-Control'] = "no-cache"
        headers['Connection'] = "keep-alive"
        me._headers = headers

        yield b'retry: 1000\n'

        while True:
            for fn, msg in self.mbx.clonePreIter(pre.qb64b, idx):
                data = bytearray("id: {}\nevent: data\ndata: ".format(fn).encode("utf-8"))
                data.extend(msg.encode("utf-8"))
                data.extend(b'\n\n')
                idx += 1
                yield data

            yield b''


class Respondant(doing.DoDoer):
    """
    Respondant processes buffer of response messages from inbound 'exn' messages and
    routes them to the appropriate mailbox.  If destination has witnesses, send response to
    one of the (randomly selected) witnesses.  Otherwise store the response in the recipients
    mailbox locally.

    """

    def __init__(self, hab, msgs=None, mbx=None, **kwa):
        """
        Creates Respondant that uses local environment to find the destination KEL and stores
        peer to peer messages in mbx, the mailboxer

        Parameters:
            hab (Habitat):  local environment
            mbx (Mailboxer): storage for local messages

        """
        self.msgs = msgs if msgs is not None else decking.Deck()

        self.hab = hab
        self.mbx = mbx if mbx is not None else exchanging.Mailboxer(name=hab.name)

        doers = [doing.doify(self.responseDo)]
        super(Respondant, self).__init__(doers=doers, **kwa)


    def responseDo(self, tymth=None, tock=0.0, **opts):
        """
        Doifiable Doist compatibile generator method to process response messages from `exn` handlers.
        If dest is not in local environment, ignore the response (for now).  If dest has witnesses,
        pick one at random and send the response to that witness for storage in the recipients mailbox
        on that witness.  Otherwise this is a peer to peer HTTP message and should be stored in a mailbox
        locally for the recipient.

        Usage:
            add result of doify on this method to doers list
        """
        while True:
            while self.msgs:
                rep = self.msgs.popleft()
                dest = rep["dest"]
                msg = rep["msg"]

                kever = self.hab.kevers[dest]
                if kever is None:
                    logger.Error("unable to reply, dest {} not found".format(dest))
                    continue

                if len(kever.wits) == 0:
                    self.mbx.storeMsg(dest=dest, msg=msg)
                else:
                    wit = random.choice(kever.wits)
                    loc = obtaining.getwitnessbyprefix(wit)

                    client = http.clienting.Client(hostname=loc.ip4, port=loc.http)
                    clientDoer = http.clienting.ClientDoer(client=client)

                    self.extend([clientDoer])

                    createCESRRequest(msg, client)

                    while not client.responses:
                        yield self.tock

                    self.remove([clientDoer])

                yield  # throttle just do one cue at a time
            yield



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.watcher module

"""
import argparse
import sys

from hio.base import doing
from hio.core.tcp import clienting

from keri.app import habbing
from keri.app.cli.commands.init import KLIRecord
from keri.db import koming
from keri.peer import exchanging

parser = argparse.ArgumentParser(description='Start watcher')
parser.set_defaults(handler=lambda args: handler())


def handler():
    startIst = StartIst(tock=0.03125)
    startIst.do()

    return


class StartIst(doing.Doist):

    def __init__(self, real=False, limit=None, doers=None, **kwa):
        self.hab = habbing.Habitat(name='kli', temp=False)
        kli = koming.Komer(db=self.hab.db, schema=KLIRecord, subkey='kli.').get((self.hab.pre,))

        super().__init__(real, limit, doers, **kwa)

        self.client = clienting.Client(tymth=self.tymen(), host='127.0.0.1', port=5678)
        clientDoer = clienting.ClientDoer(client=self.client)

        self.extend([clientDoer, doing.doify(self.infoDoer)])


    def infoDoer(self, tymth=None, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        while not self.client.connected:
            (yield self.tock)

        print("connected")
        payload = dict(
            cmd='watcher',
            args=('start',),
        )

        srdr = exchanging.exchange(route="/cmd/", payload=payload)
        excMsg = self.hab.sanction(srdr)

        self.client.tx(excMsg)
        self.client.close()

        print("watcher start command sent")
        sys.exit(0)



# -*- encoding: utf-8 -*-
"""
KERI
keri.demo.demoing module

Utilities for demos
"""
import argparse
import logging
import os

from hio.core import wiring
from hio.core.tcp import clienting

from keri import __version__
from keri import help
from keri.app import habbing, keeping, directing
from keri.db import basing
from keri.demo.demoing import IanDirector
from keri.vdr import issuing

logger = help.ogler.getLogger()


def runDemo(did, lei, witness=5621, peer=5629, expire=0.0):
    """
    Setup and run one demo controller for sam, like bob only better
    """

    secrets = [
        'ArwXoACJgOleVZ2PY7kXn7rA0II0mHYDhc6WrBH8fDAc',
        'A6zz7M08-HQSFq92sJ8KJOT2cZ47x7pXFQLPB0pckB3Q',
        'AcwFTk-wgk3ZT2buPRIbK-zxgPx-TKbaegQvPEivN90Y',
        'Alntkt3u6dDgiQxTATr01dy8M72uuaZEf9eTdM-70Gk8',
        'A1-QxDkso9-MR1A8rZz_Naw6fgaAtayda8hrbkRVVu1E',
        'AKuYMe09COczwf2nIoD5AE119n7GLFOVFlNLxZcKuswc',
        'AxFfJTcSuEE11FINfXMqWttkZGnUZ8KaREhrnyAXTsjw',
        'ALq-w1UKkdrppwZzGTtz4PWYEeWm0-sDHzOv5sq96xJY'
    ]

    doers = setupController(secrets=secrets,
                            did=did,
                            lei=lei,
                            witnessPort=witness,
                            peerPort=peer)

    directing.runController(doers=doers, expire=expire)


def setupController(secrets, did, lei, witnessPort=5621, peerPort=5629, indirect=False):
    """
    Setup and return doers list to run controller
    """
    name = "ian"

    secrecies = []
    for secret in secrets:  # convert secrets to secrecies
        secrecies.append([secret])

    # setup databases for dependency injection
    ks = keeping.Keeper(name=name, temp=True)  # not opened by default, doer opens
    ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
    db = basing.Baser(name=name, temp=True)  # not opened by default, doer opens
    dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

    # setup habitat
    hab = habbing.Habitat(name=name, ks=ks, db=db, temp=True, secrecies=secrecies)
    habDoer = habbing.HabitatDoer(habitat=hab)  # setup doer

    iss = issuing.Issuer(hab=hab, name=name, noBackers=True)
    issDoer = issuing.IssuerDoer(issuer=iss)

    # setup doers
    regDoer = basing.BaserDoer(baser=iss.reger)

    path = os.path.dirname(__file__)
    path = os.path.join(path, 'logs')

    wl = wiring.WireLog(samed=True, filed=True, name=name, prefix='demo', reopen=True,
                        headDirPath=path)
    wireDoer = wiring.WireLogDoer(wl=wl)

    witnessClient = clienting.Client(host='127.0.0.1', port=witnessPort, wl=wl)
    witnessClientDoer = clienting.ClientDoer(client=witnessClient)

    peerClient = clienting.Client(host='127.0.0.1', port=peerPort, wl=wl)
    peerClientDoer = clienting.ClientDoer(client=peerClient)

    director = IanDirector(hab=hab,
                           issuer=iss,
                           witnessClient=witnessClient,
                           peerClient=peerClient,
                           tock=0.125,
                           recipientIdentifier=did,
                           lei=lei)

    reactor = directing.Reactor(hab=hab, client=witnessClient, indirect=indirect)
    peerReactor = directing.Reactor(hab=hab, client=peerClient, indirect=indirect)

    logger.info("\nDirect Mode demo of %s:\nNamed %s to TCP port %s.\n\n",
                hab.pre, hab.name, witnessPort)

    return [ksDoer, dbDoer, habDoer, issDoer, regDoer, wireDoer, witnessClientDoer, director,
            reactor, peerClientDoer, peerReactor]



def parseArgs(version=__version__):
    d = "Runs KERI direct mode demo controller.\n"
    d += "Example:\nkeri_vic -r 5621 -l 5620 --e 10.0'\n"
    p = argparse.ArgumentParser(description=d)
    p.add_argument('-V', '--version',
                   action='version',
                   version=version,
                   help="Prints out version of script runner.")
    p.add_argument('-w', '--witness',
                   action='store',
                   default=5621,
                   help="Remote witness port number the witness client connects to. Default is 5621.")
    p.add_argument('-p', '--peer',
                   action='store',
                   default=5629,
                   help="Remote peer port number the peer client connects to. Default is 5629.")
    p.add_argument('-e', '--expire',
                   action='store',
                   default=0.0,
                   help="Expire time for demo. 0.0 means not expire. Default is 0.0.")
    p.add_argument('-d', '--did',
                   action='store',
                   help="Recipient DID method specific identifier (Required)")
    p.add_argument('-l', '--lei',
                   action='store',
                   help="Legal Entity Identifier (Required)")

    args = p.parse_args()

    return args


def main():
    args = parseArgs(version=__version__)

    help.ogler.level = logging.INFO
    help.ogler.reopen(name="ian", temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Demo for Ian connecting to "
                "%s.******\n\n", args.peer)

    runDemo(did=args.did,
            lei=args.lei,
            witness=int(args.witness),
            peer=int(args.peer),
            expire=args.expire)

    logger.info("\n******* Ended Demo for Ian connecting to "
                "%s.******\n\n", args.peer)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.witness module

"""

import argparse

parser = argparse.ArgumentParser(description='Stop witness')
parser.set_defaults(handler=lambda args: handler())


def handler():
    print('witness stop')



# -*- encoding: utf-8 -*-
"""
keri.peer.exchanging module

"""

from datetime import timedelta

from hio.base import doing
from hio.help import decking

from .. import help
from ..core import eventing, coring
from ..core.coring import MtrDex
from ..db import subing, dbing
from ..help import helping
from ..kering import ValidationError, MissingSignatureError, MissingDestinationError, AuthZError, ExchangeError

ExchangeMessageTimeWindow = timedelta(seconds=1010000)

logger = help.ogler.getLogger()


class Exchanger(doing.DoDoer):
    """
     Peer to Peer KERI message Exchanger.
    """

    def __init__(self, hab, handlers, controller=None, cues=None, delta=ExchangeMessageTimeWindow, **kwa):
        """
        Initialize instance

        Parameters:
            hab (Habitat): instance of local controller's context
            cues (Deck):  of Cues i.e. notices of requests needing response
            delta (timedelta): message timeout window
        """

        self.hab = hab
        self.controller = controller
        self.kevers = hab.kvy.kevers
        self.delta = delta
        self.routes = dict()
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque

        doers = []
        for handler in handlers:
            if handler.resource in self.routes:
                raise ValidationError("unable to register behavior {}, it has already been registered"
                                      "".format(handler.resource))

            self.routes[handler.resource] = handler
            doers.append(handler)

        super(Exchanger, self).__init__(doers=doers, **kwa)

    def processEvent(self, serder, source, sigers):
        """
        Process one serder event with attached indexed signatures representing
        a Peer to Peer exchange message.

        Parameters:
            serder (Serder) instance of event to process
            source (Prefixer) identifier prefix of event sender
            sigers (list) of Siger instances of attached controller indexed sigs

        """

        route = serder.ked["r"]
        payload = serder.ked["d"]
        dts = serder.ked["dt"]

        if route not in self.routes:
            raise AttributeError("unregistered route {} for exchange message = {}"
                                 "".format(route, serder.pretty()))

        behavior = self.routes[route]

        # delta = behavior.delta if behavior.delta is not None else self.delta
        delta = self.delta
        msgDt = helping.fromIso8601(dts)
        now = helping.nowUTC()

        if now - msgDt > delta:
            raise ValidationError("message received outside time window with delta {} message={}"
                                  "".format(delta, serder.pretty()))

        sever = self.kevers[source.qb64]

        if self.controller is not None and self.controller != source.qb64:
            raise AuthZError("Message {} is from invalid source {}"
                             "".format(payload, source.qb64))

        #  Verify provided sigers using verfers
        sigers, indices = eventing.verifySigs(serder=serder, sigers=sigers, verfers=sever.verfers)
        if not sever.tholder.satisfy(indices):  # at least one but not enough
            self.escrowPSEvent(serder=serder, sigers=sigers)
            raise MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                        " for evt = {}.".format(sever.tholder.sith,
                                                                [siger.qb64 for siger in sigers],
                                                                serder.ked))

        msg = dict(
            payload=payload,
            pre=source,
            sigers=sigers,
            verfers=sever.verfers,
        )

        behavior.msgs.append(msg)

    def processResponseIter(self):
        """
        Iterate through cues and yields one or more responses for each cue.

        Parameters:
            cues is deque of cues

        """
        responses = []
        for _, behavior in self.routes.items():  # get responses from all behaviors
            while behavior.cues:
                excSrdr = behavior.cues.popleft()
                msg = self.hab.sanction(excSrdr)
                responses.append(dict(dest=excSrdr.ked["i"], msg=msg))

        while responses:  # iteratively process each response in responses
            msg = responses.pop(0)
            yield msg

    def escrowPSEvent(self, serder, sigers):
        """
        Escrow event that does not have enough signatures.

        Parameters:
            serder is Serder instance of event
            sigers is list of Siger instances of indexed controller sigs
        """
        pass


def exchange(route, payload, recipient=None, date=None, modifiers=None, version=coring.Version,
             kind=coring.Serials.json):
    """
    Create an `exn` message with the specified route and payload
    Parameters:
        route (string) to destination route of the message
        payload (dict) body of message to deliver to route
        recipient (str) qb64 identifier prefix of target of message
        date (str) Iso8601 formatted date string to use for this request
        modifiers (dict) equivalent of query string of uri, modifiers for the request that are not
                         part of the payload
        version (Version) is Version instance
        kind (Serials) is serialization kind

    """
    vs = coring.Versify(version=version, kind=kind, size=0)
    ilk = eventing.Ilks.exn
    dt = date if date is not None else helping.nowIso8601()

    ked = dict(v=vs,
               t=ilk,
               i=recipient,
               dt=dt,
               r=route,
               d=payload,
               q=modifiers
               )

    if recipient is None:
        del ked["i"]

    if modifiers is None:
        del ked["q"]

    return eventing.Serder(ked=ked)  # return serialized ked


class StoreExchanger:
    """
    StoreExchanger receives exn messages and stores them indexed by prefix identifier of the target.

    """

    def __init__(self, hab, mbx=None, exc=None):
        self.hab = hab
        self.kevers = self.hab.kevers
        self.db = mbx if mbx is not None else Mailboxer(name=hab.name)
        self.exc = exc if exc is not None else Exchanger(hab=hab, handlers=[])


    @property
    def routes(self):
        return self.exc.routes

    def processEvent(self, serder, source, sigers):
        """
        Process one serder event with attached indexed signatures representing
        a Peer to Peer exchange message.

        Parameters:
            serder (Serder) instance of event to process
            source (Prefixer) identifier prefix of event sender
            sigers (list) of Siger instances of attached controller indexed sigs

        """

        sever = self.kevers[source.qb64]

        #  Verify provided sigers using verfers
        sigers, indices = eventing.verifySigs(serder=serder, sigers=sigers, verfers=sever.verfers)
        if not sever.tholder.satisfy(indices):  # at least one but not enough
            self.escrowPSEvent(serder=serder, source=source, sigers=sigers)
            raise MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                        " for evt = {}.".format(sever.tholder.sith,
                                                                [siger.qb64 for siger in sigers],
                                                                serder.ked))
        if "i" not in serder.ked:
            raise MissingDestinationError("Failure saving evt = {} from = {} in mailbox, missing destination"
                                          "".format(serder.ked, source.qb64))

        if serder.pre == self.hab.pre:
            self.exc.processEvent(serder=serder, source=source, sigers=sigers)
        elif self.hab.pre in sever.wits:
            dest = eventing.Prefixer(qb64=serder.pre)
            msg = self._reconstruct(serder=serder, source=source, sigers=sigers)
            self.db.storeMsg(dest=dest.qb64b, msg=msg)
        else:
            raise ExchangeError("Event recipient {} is neither this witness {} nor a prefix for whom this"
                                "witness is a witness {}".format(serder.pre, self.hab.pre, sever.wits))

    def processResponseIter(self):
        yield from self.exc.processResponseIter()

    def escrowPSEvent(self, serder, source, sigers):
        """
        Escrow event that does not have enough signatures.

        Parameters:
            serder is Serder instance of event
            sigers is list of Siger instances of indexed controller sigs
        """
        pass


    @staticmethod
    def _reconstruct(serder, source, sigers):
        msg = bytearray()  # message

        msg.extend(serder.raw)

        msg.extend(coring.Counter(coring.CtrDex.SignerSealCouples, count=1).qb64b)
        msg.extend(source.qb64b)

        if sigers:
            msg.extend(coring.Counter(code=coring.CtrDex.ControllerIdxSigs, count=len(sigers)).qb64b)

            for sig in sigers:
                msg.extend(sig.qb64b)
        return msg



class Mailboxer(dbing.LMDBer):
    """
    Mailboxer stores exn messages in order and provider iterator access at an index.


    """
    TailDirPath = "keri/mbx"
    AltTailDirPath = ".keri/mbx"
    TempPrefix = "keri_mbx_"

    def __init__(self, name="mbx", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            dirMode:
            reopen:
            kwa:
        """
        self.fels = None
        self.msgs = None

        super(Mailboxer, self).__init__(name=name, headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """

        :param kwa:
        :return:
        """
        super(Mailboxer, self).reopen(**kwa)

        self.fels = self.env.open_db(key=b'fels.')
        self.msgs = subing.Suber(db=self, subkey='msgs.')  # key states

        return self.env

    def delFe(self, key):
        """
        Use snKey()
        Deletes value at key.
        Returns True If key exists in database Else False
        """
        return self.delVal(self.fels, key)

    def appendFe(self, pre, val):
        """
        Return first seen order number int, fn, of appended entry.
        Computes fn as next fn after last entry.
        Uses fnKey(pre, fn) for entries.

        Append val to end of db entries with same pre but with fn incremented by
        1 relative to last preexisting entry at pre.

        Parameters:
            pre is bytes identifier prefix for event
            val is event digest
        """
        return self.appendOrdValPre(db=self.fels, pre=pre, val=val)

    def _getFelItemPreIter(self, pre, fn=0):
        """
        Returns iterator of all (fn, dig) duples in first seen order for all events
        with same prefix, pre, in database. Items are sorted by fnKey(pre, fn)
        where fn is first seen order number int.
        Returns a First Seen Event Log FEL.
        Returned items are duples of (fn, dig): Where fn is first seen order
        number int and dig is event digest for lookup in .evts sub db.

        Raises StopIteration Error when empty.

        Parameters:
            pre is bytes of itdentifier prefix
            fn is int fn to resume replay. Earliset is fn=0
        """
        return self.getAllOrdItemPreIter(db=self.fels, pre=pre, on=fn)

    def getFelItemAllPreIter(self, key=b''):
        """
        Returns iterator of all (pre, fn, dig) triples in first seen order for
        all events for all prefixes in database. Items are sorted by
        fnKey(pre, fn) where fn is first seen order number int.
        Returns all First Seen Event Logs FELs.
        Returned items are tripes of (pre, fn, dig): Where pre is identifier prefix,
        fn is first seen order number int and dig is event digest for lookup
        in .evts sub db.

        Raises StopIteration Error when empty.

        Parameters:
            key is key location in db to resume replay, If empty then start at
                first key in database
        """
        return self.getAllOrdItemAllPreIter(db=self.fels, key=key)


    def storeMsg(self, dest, msg):
        """
        Add exn event to mailbox of dest identifier

        Parameters:
            msg (bytes):
            dest (qb64b):

        """
        if hasattr(dest, "encode"):
            dest = dest.encode("utf-8")

        if hasattr(msg, "encode"):
            msg = msg.encode("utf-8")

        digb = coring.Diger(ser=msg, code=MtrDex.Blake3_256).qb64b
        self.appendFe(dest, digb)
        self.msgs.pin(keys=digb, val=msg)


    def clonePreIter(self, pre, fn=0):
        """
        Returns iterator of first seen exn messages with attachments for the
        identifier prefix pre starting at first seen order number, fn.

        """
        if hasattr(pre, 'encode'):
            pre = pre.encode("utf-8")

        for fn, dig in self._getFelItemPreIter(pre, fn=fn):
            if msg := self.msgs.get(keys=dig):
                yield fn, msg




# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import os
import sys

from hio import help
from hio.base import doing

from keri.app.cli.serving import Serving

logger = help.ogler.getLogger()


def main():
    # do the UNIX double-fork magic, see Stevens' "Advanced
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError as e:
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent, print eventual PID before
            print("Daemon PID", pid)
            sys.exit(0)
    except OSError as e:
        sys.exit(1)

    print("starting serving...")

    doist = doing.Doist(tock=0.03125, real=True)
    serving = Serving(tymth=doist.tymen())
    doist.do(doers=[serving])


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
"""
keri.vc.handling module

"""
from hio.base import doing
from hio.help import decking

from .. import help
from ..core.coring import dumps, Deversify
from ..core.scheming import JSONSchema
from ..kering import ShortageError
from ..peer import exchanging
from ..vc.proving import Credentialer
from ..vc.walleting import parseCredential, buildProof

logger = help.ogler.getLogger()


class OfferHandler(doing.Doer):
    """
    Sample handler to a credential offer message from an Issuer to a Holder for a credential based
    on
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/offer"                                // resource with nested namespace
           "q": {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "output_descriptors: [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                "cesr": {
                  "proof_type": ["Ed25519Signature2018"]
                }
              }
           } //embedded credential_manifest like structure, may contain presentation_definition of requirements for fullfilment
        }-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE-d_ApsFEgCA

    """

    resource = "/credential/offer"

    def __init__(self, wallet, formats, typ=JSONSchema(), cues=None, **kwa):
        """

        Parameters:
            wallet (Wallet) credential wallet that will hold the issued credentials
            formats (list) of format str names accepted for offers
            typ (JSONSchema) credential type to accept
        """
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.wallet = wallet
        self.formats = formats
        self.typ = typ

        super(OfferHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]

                issuer = payload["issuer"]
                descriptors = payload["input_descriptors"]
                formats = payload["format"]

                schema = descriptors[0]

                fmts = []
                for fmt in self.formats:
                    if fmt in formats:
                        fmts.append(formats[fmt])

                if not fmts:
                    logger.info("No acceptable formats being offered in {}.  Needed one of {}."
                                "".format(formats, self.formats))
                    continue

                apply = credential_apply(issuer, schema, format)

                self.cues.append(exchanging.exchange(route="/credential/apply", payload=apply))

                yield

            yield


class ApplyHandler(doing.DoDoer):
    """
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/apply"
           "q" {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "input_descriptors": [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                 "cesr": {
                   "proof_type": ["Ed25519Signature2018"]
                 }
              }
           } //embedded credential_submission, may contain credential_fullfilment responding to presentation_def above
        }-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE-d_ApsFEgCA

    """

    resource = "/credential/apply"

    def __init__(self, hab, typ=JSONSchema(), cues=None, **kwa):
        """

        Parameters:
            hab (Habitat) credential wallet that will hold the issued credentials
            typ (JSONSchema) credential type to accept
        """
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.hab = hab
        self.typ = typ

        super(ApplyHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                logger.info(payload)
                yield

            yield


class IssueHandler(doing.Doer):
    """
    Sample class that handles a credential Issue `exn` message.  By default, this handler
    stores the credential in the provided wallet.  The incoming message must have the following format:

         {
       "vc" [
         {
           "vc": {
              "v": "KERI10JSON00011c_", //KERI Version String
              "x": "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI", // Identifier prefix of the Schema
              "d": {
                   "type": [
                       "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI"
                   ],
                   "id": "did:keri:EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI",
                   "issuer": "did:keri:EchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA",
                   "issuanceDate": "2021-06-09T17:35:54.169967+00:00",
                   "credentialSubject": {
                       "id": "did:keri:did:keri:Efaavv0oadfghasdfn443fhbyyr4v",
                       "lei": "254900OPPU84GM83MG36"
                   },
                   "credentialSchema": {
                       "id": ""
                       "type": ""
                   },
                   "credentialStatus": {
                      "id": "",
                      "type": ""
                   }
              }
           }, // embedded verifiable credential
           "proof": "-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE
                 -d_ApsFEgCA-GAB0AAAAAAAAAAAAAAAAAAAAABQEchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA"
           }
       ]   //list of verifiable credentials
    }


    """

    resource = "/credential/issue"

    def __init__(self, wallet, typ=JSONSchema(), cues=None, **kwa):
        """

        Parameters:
            wallet (Wallet) credential wallet that will hold the issued credentials
            typ (JSONSchema) credential type to accept
        """
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.wallet = wallet
        self.typ = typ

        super(IssueHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                envelopes = payload["vc"]
                for envlop in envelopes:
                    crd = envlop["vc"]
                    proof = envlop["proof"]

                    vs = crd["v"]

                    kind, version, size = Deversify(vs)
                    raw = dumps(ked=crd, kind=kind)
                    if len(raw) != size:
                        raise ValueError("invalid length {} for credential = {}".format(size, crd))

                    msg = bytearray(raw)
                    msg.extend(proof.encode("utf-8"))

                    parseCredential(ims=msg, wallet=self.wallet, typ=self.typ)

                yield

            yield


class RequestHandler(doing.Doer):
    """
        Processor for a credential request with input descriptors in the payload used to
        match saved credentials based on a schema.  The payload of the request is expected to
        have the following format:

             {
                "submission_requirements": [{
                   "name": "Proof of LEI",
                   "rule": "pick",
                   "count": 1,
                   "from": "A"
                 }]
                "input_descriptors": [
                   {
                      "x":"EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA",
                      "group": ["A"],
                   }
                ],
                "format": {
                   "cesr": {
                     "proof_type": ["Ed25519Signature2018"]
                   }
                }
             }

    """

    resource = "/presentation/request"

    def __init__(self, wallet, typ=JSONSchema(), cues=None, **kwa):
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.wallet = wallet
        self.typ = typ

        super(RequestHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """
        Process presentation request message with sender identifier, sigs and verfers

        Messages:
            payload is dict representing the body of a /presentation/request message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /presentation/request message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                requestor = msg["pre"]
                descriptors = payload["input_descriptors"]

                matches = []
                for descriptor in descriptors:
                    said = descriptor["x"]
                    credentials = self.wallet.getCredentials(said)
                    if len(credentials) > 0:
                        vc = credentials[0][0].pretty()
                        # logger.info("Presenting Credential for schema %s:\n VC=%s", said, vc)
                        matches.append(credentials[0])

                if len(matches) > 0:
                    pe = presentation_exchange(matches)
                    self.cues.append(exchanging.exchange(route="/presentation/proof", payload=pe,
                                                         recipient=requestor.qb64))

                yield

            yield


class ProofHandler(doing.Doer):
    """
    Processor for responding to presentation proof peer to peer message.  The payload of the message
    is expected to have the following format:

        {
          "presentation_submission": {
              "id": "a30e3b91-fb77-4d22-95fa-871689c322e2",
              "definition_id": "32f54163-7166-48f1-93d8-ff217bdb0653",
              "descriptor_map": [
                   {
                     "id": "gleif_vlei",
                     "format": "cesr",
                     "path": "$.verifiableCredential[0]"
                   }
              ]
          },
          "verifiableCredential": [
              {
                "vc": {
                   "v": "KERI10JSON00011c_", //KERI Version String
                   "x": "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI", // Identifier prefix of the Schema
                   "q": {
                        "type": [
                            "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI"
                        ],
                        "id": "did:keri:EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI",
                        "issuer": "did:keri:EchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA",  //Identifier prefix of
                        the issuer
                        "issuanceDate": "2021-06-09T17:35:54.169967+00:00",
                        "credentialSubject": {
                            "id": "did:keri:did:keri:Efaavv0oadfghasdfn443fhbyyr4v",
                            "lei": "254900OPPU84GM83MG36"
                        }
                    } // embedded verifiable credential
                }
                "proof": "-VA0-FABE4YPqsEOaPNaZxVIbY-Gx2bJgP-c7AH_K7pEE
                -YfcI9E0AAAAAAAAAAAAAAAAAAAAAAAElHzHwX3V6itsD2Ksg_CNBbUNTBYzLYw-AxDNI7_ZmaI
                -AABAALK_6pkUjCx76CTE7pNrKlhC84ewgZbs-4ciTvLcxkwUaWX7ukpr55P9RtJhlHxdtz3kN0zfM0HrtFitKuPy3BA"
              }
           ]
        }
    """

    resource = "/presentation/proof"

    def __init__(self, typ=JSONSchema(), cues=None, proofs=None, **kwa):
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.proofs = proofs if proofs is not None else decking.Deck()

        self.typ = typ

        super(ProofHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                pre = msg["pre"]

                if "presentation_submission" not in payload:
                    raise ValueError("invalid presentation proof payload")

                if "verifiableCredential" not in payload:
                    raise ValueError("invalid presentation proof payload")

                pe = payload["presentation_submission"]
                vcs = payload["verifiableCredential"]

                if "descriptor_map" not in pe:
                    raise ValueError("invalud presentation submission in proof payload")

                dm = pe["descriptor_map"]

                for idx, descriptor in enumerate(dm):
                    # TODO:  Find verifiable credential in vcs based on `path`
                    vc = vcs[idx]
                    self.proofs.append((pre, vc))

                yield

            yield


def envelope(msg, typ=JSONSchema()):
    """
    Returns a dict of a VC split into the "vc" and "proof"

    Parameters:
        msg: bytes of verifiable credential to split
        typ: schema type of the VC
    """

    ims = bytearray(msg)
    try:
        creder = Credentialer(raw=ims, typ=typ)
    except ShortageError as e:
        raise e
    else:
        del ims[:creder.size]

    return dict(
        vc=creder.crd,
        proof=ims.decode("utf-8")
    )


def presentation_exchange(credentials):
    dm = []
    vcs = []

    for idx, (creder, prefixer, seqner, diger, sigers) in enumerate(credentials):
        proof = buildProof(prefixer, seqner, diger, sigers)
        dm.append(dict(
            id=creder.schema,
            format="cesr",
            path="$.verifiableCredential[{}]".format(idx)
        ))

        vcs.append(dict(
            vc=creder.crd,
            proof=proof.decode("utf-8")
        ))

    d = dict(
        presentation_submission=dict(
            descriptor_map=dm
        ),
        verifiableCredential=vcs
    )

    return d


def credential_apply(issuer, schema, formats):
    """
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/apply"
           "q" {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "input_descriptors": [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                 "cesr": {
                   "proof_type": ["Ed25519Signature2018"]
                 }
              }
           } //embedded credential_submission, may contain credential_fullfilment responding to presentation_def above
        }

    Parameters:
        issuer (str) is qb64 identifier prefix of the issuer
        schema (str) is qb64 SAID of schema being applied for
        formats (list of CredentialFormat) is list of acceptable credential formats

    """

    d = dict(
        issuer=issuer,
        input_descriptors=[schema],
        format=[]
    )

    for fmt in formats:
        d["format"].append(fmt.fmd)

    return d


def credential_issue(msgs, typ):
    """
    Returns a list of credentials enveloped inside a credential issue message

    Parameters:
        msgs (list) is list of CESR formatted, endorsed verifiable credentials
        typ (JSONSchema) type of credentials being issued

    """

    vcs = []
    for msg in msgs:
        vc = envelope(msg, typ)
        vcs.append(vc)

    pl = dict(
        vc=vcs
    )

    return pl



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from hio.base import doing

from keri.app import habbing
from keri.vdr.issuing import Issuer

parser = argparse.ArgumentParser(description='Revoke a verifiable credential')
parser.set_defaults(handler=lambda args: RevokeDoer(vcdig=args.vcdig, hab=args.hab))
parser.add_argument('--vcdig', help='vcdig is hash digest of vc content qb64')


class RevokeDoer(doing.Doer):

    def __init__(self, vcdig, tock=0.0, hab: habbing.Habitat = None, **kwa):
        self.hab = hab
        self.vcdig = vcdig
        super(RevokeDoer, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        iss = Issuer(hab=self.hab, name=self.hab.name)

        iss.revoke(vcdig=self.vcdig)

        print(f"Rotated keys for {self.hab.name}")
        print(f"New public key {self.hab.kever.verfers[0].qb64}")

        return super().do(tymth, tock, **opts)

